#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

static unsigned long pc, dpc;  
static unsigned long reg[1 << 6];
static unsigned long mem[1 << 20];
static FILE *fpx = NULL, *fpr = NULL, *fpw = NULL;
static unsigned long inst;
static char *status = NULL;

struct quasifloat {
  unsigned short s;
  unsigned short e;
  unsigned long f;
};

/* interpretation */

static unsigned long op(void) {
  return inst >> 28 & 0xFUL;
}

static unsigned long rs(void) {
  return inst >> 22 & 0x3FUL;
}

static unsigned long rt(void) {
  return inst >> 16 & 0x3FUL;
}

static unsigned long rd(void) {

  return inst >> 10 & 0x3FUL;
}

static unsigned long shft(void) {
  return inst >> 4 & 0x1FUL;
}

static unsigned long fnct(void) {
  return inst & 0xFUL;
}

static unsigned long opt(void) {
  return inst >> 16 & 0x1FUL | (inst & 0x200000UL ? 0xFFFFFFE0UL : 0UL);
}

static unsigned long oft(void) {
  return inst & 0x7FFFUL | (inst & 0x8000UL ? 0xFFFF8000UL : 0UL);
}

/* conversion between long and quasifloat */

static struct quasifloat split(const unsigned long l) {
  /* convert long into quasifloat */
  const struct quasifloat qf = {
    l >> 31,
    l >> 23 & 0xFF,
    (l & 0x7F800000UL ?  0x800000UL : 0UL) | l & 0x7FFFFFUL
  };
  return qf;
}

static long join(const struct quasifloat qf) {
  unsigned short z;
  /* count leading zero(s) */
  for (z = 0; z < 25; z++) {
    if (qf.f << z & 0x1000000UL) break;
  }  
  /* convert quasifloat into long */
  return  qf.e - z < 0xFFUL && z != 25 ?
    qf.s << 31 | qf.e - z + 1UL << 23 & 0x7F800000UL | qf.f << z >> 1 & 0x7FFFFFUL :
    0UL;
}

/* interpretation */

static void add(void) {
  reg[rd()] = reg[rs()] + reg[rt()] & 0xFFFFFFFFUL;
}

static void sub(void) {
  reg[rd()] = reg[rs()] - reg[rt()] & 0xFFFFFFFFUL;
}

static void sll(void) {
  reg[rd()] = reg[rs()] << shft() & 0xFFFFFFFFUL;
}

static void srr(void) {
  reg[rd()] = reg[rs()] & 0x80000000 ? ~(~reg[rs()] >> shft()) : reg[rs()] >> shft();
}

static void and(void) {
  reg[rd()] = reg[rs()] & reg[rt()];
}

static void or(void) {
  reg[rd()] = reg[rs()] | reg[rt()];
}

static void xor(void) {
  reg[rd()] = reg[rs()] ^ reg[rt()];
}

static void xnor(void) {
  reg[rd()] = ~reg[rs()] & 0xFFFFFFFFUL ^ reg[rt()];
}

static void fadd(void) {
  struct quasifloat lqf = split(reg[rs()]);
  struct quasifloat rqf = split(reg[rt()]);
  if (lqf.e > rqf.e || lqf.e == rqf.e && lqf.f >= rqf.f) {
    lqf.f = lqf.e - rqf.e > 23 ?
      lqf.f :
      lqf.s ^ rqf.s ?
      lqf.f - (rqf.f >> lqf.e - rqf.e) :
      lqf.f + (rqf.f >> lqf.e - rqf.e);
    reg[rd()] = join(lqf);
  } else {
    rqf.f = rqf.e - lqf.e > 23 ?
      rqf.f :
      lqf.s ^ rqf.s ?
      rqf.f - (lqf.f >> rqf.e - lqf.e) :
      rqf.f + (lqf.f >> rqf.e - lqf.e);
    reg[rd()] = join(rqf);
  }
}

static void fsub(void) {
  struct quasifloat lqf = split(reg[rs()]);
  struct quasifloat rqf = split(reg[rt()]);
  if (lqf.e > rqf.e
      || lqf.e == rqf.e && lqf.f >= rqf.f) {
    lqf.f = lqf.e - rqf.e > 23 ?
      lqf.f :
      lqf.s ^ rqf.s ?
      lqf.f + (rqf.f >> lqf.e - rqf.e):
      lqf.f - (rqf.f >> lqf.e - rqf.e);
    reg[rd()] = join(lqf);
  } else {
    rqf.f = rqf.e - lqf.e > 23 ?
      rqf.f :
      lqf.s ^ rqf.s ?
      rqf.f + (lqf.f >> rqf.e - lqf.e) :
      rqf.f - (lqf.f >> rqf.e - lqf.e);
    rqf.s = !rqf.s;
    reg[rd()] = join(rqf);
  }
}

static void fmul(void) {
  struct quasifloat lqf = split(reg[rs()]);
  struct quasifloat rqf = split(reg[rt()]);
  if (lqf.f & 0x800000 && rqf.f & 0x800000) {
    lqf.f = (long long)(lqf.f / 0x1000UL) * (rqf.f / 0x1000UL) * 0x2UL
      + (long long)(lqf.f / 0x1000UL) * (rqf.f % 0x1000UL) / 0x800UL
      + (long long)(lqf.f % 0x1000UL) * (rqf.f / 0x1000UL) / 0x800UL
      + 0x2UL;
  } else {
    lqf.f = 0UL;
  }
  lqf.e += rqf.e - 0x80;
  lqf.s ^= rqf.s;
  reg[rd()] = join(lqf);
}

static void finv(void) {
  struct quasifloat qf = split(reg[rs()]);
  const unsigned long addr
    = qf.f >> 13 & 0x3FEUL | 0x401UL;
  const unsigned long long tbl
    = 0x400000ULL / addr - 0x200000ULL / addr & 0x3FFUL | 0x400UL;
  qf.f = tbl * (0x1000003UL - qf.f / 2 * tbl / 0x400UL) / 0x400UL + 1;
  qf.e = 0xFF - qf.e;
  reg[rd()] = join(qf);
}

static void fsqrt(void) {
  struct quasifloat qf = split(reg[rs()]);
  if (qf.f & 0x800000) {
    const unsigned long addr
      = (qf.f >> 14 & 0x1FEUL | 0x201UL) << (qf.e & 0x1UL);
    const unsigned long long itbl
      = (int)(1. / sqrt((double)addr / 0x200UL) * 0x1000UL + 0.5) >> 1 & 0x3FF | 0x400;
    const unsigned long long tbl
      = 0x100000000ULL / itbl - 0x80000000ULL / itbl - 1 & 0x0FFFFFUL | 0x100000;
    qf.f = tbl * 0x4UL
      + (qf.f / 2 * itbl / 0x800 << qf.e % 2);
    qf.e = qf.e / 2 + 64;
    reg[rd()] = join(qf);
  } else {
    reg[rd()] = 0UL;
  }
}

static void get(void) {
  if ((reg[rd()] = fgetc(fpr)) == (unsigned)EOF) {
    status = "Message: USB exception detected\n";
  }
}

static void put(void) {
  fputc(reg[rs()], fpw);
}

static void fnct_error(void) {
  status = "Message: Undefined function detected\n";
}

static const char *fnct_label[16] = {
  "ADD", "SUB", "SLL", "SRR",
  "AND", "OR", "XOR", "XNOR",
  "FADD", "FSUB", "FMUL", NULL,
  "FINV", "FSQRT", "GET", "PUT",
};

static void (* fnct_behavior[16])(void) = {
  add, sub, sll, srr,
  and, or, xor, xnor,
  fadd, fsub, fmul, fnct_error,
  finv, fsqrt, get, put,
};

static long long fnct_count[16];

static unsigned long lcache_tbl[1 << 12];
static unsigned long long lcache_count;

static void calc(void) {
  const unsigned long addr = fnct();
  fnct_count[addr]++;
  fnct_behavior[addr]();
}

static void addi(void) {
  reg[rt()] = reg[rs()] + oft() & 0xFFFFFFFF;  
}

static void load(void) {
  const unsigned long size = sizeof(mem) / sizeof(unsigned long);
  const unsigned long addr = (reg[rs()] + oft()) % size;
  reg[rt()] = mem[addr];
  if (lcache_tbl[addr & 0xFFF] != addr) {
    lcache_count++;
    lcache_tbl[addr & 0xFFF] = addr;
  }
}

static void store(void) {
  const unsigned long size = sizeof(mem) / sizeof(unsigned long);
  const unsigned long addr = (reg[rs()] + oft()) % size;
  if (addr < 4
      || addr >= 0x40000 && addr < reg[1] % size - 4
      || addr >= reg[2] % size + 4) {
    //status =  "Message: SRAM exception detected.\n";
  }
#ifdef DUMP
  fprintf(stderr, "%ld %08lX\r\n", addr, reg[rt()]);
#endif
  mem[addr] = reg[rt()];
  lcache_tbl[addr & 0xFFF] = addr;
}

static void be(void) {
  if ((reg[rs()] ^ 0x80000000UL) == (opt() ^ 0x80000000)) dpc += oft();
}

static void bne(void) {
  if ((reg[rs()] ^ 0x80000000UL) != (opt() ^ 0x80000000UL)) dpc += oft();
}

static void bl(void) {
  if ((reg[rs()] | 0x80000000UL) < (opt() | 0x80000000UL)) dpc += oft();
}

static void bge(void) {
  if ((reg[rs()] | 0x80000000UL) >= (opt() | 0x80000000UL)) dpc += oft();
}

static void fbl(void) {
  if ((reg[rs()] | 0x80000000UL) < (reg[rt()] ^ 0x80000000UL)
      || (reg[rs()] ^ 0x80000000UL) >= (reg[rt()] | 0x80000000UL)) dpc += oft();
}

static void fbge(void) {
  if ((reg[rs()] | 0x80000000UL) >= (reg[rt()] ^ 0x80000000UL)
      || (reg[rs()] ^ 0x80000000UL) < (reg[rt()] | 0x80000000UL)) dpc += oft();
}

static void jr(void) {
  dpc = reg[rs()] + oft() - pc;
}

static void jal(void) {
  reg[rt()] = pc + 1;
  dpc += oft();
}

static void op_error(void) {
  status = "Message: Undefined operation detected\n";
}

static void (*op_behavior[16])() = {
  addi, calc, load, store,
  op_error, op_error, op_error, op_error,
  be, bne, bl, bge,
  fbl, fbge, jal, jr 
};

static const char *op_label[16] = {
  "CALC", "ADDI", "LOAD", "STORE",
  NULL, NULL, NULL, NULL,
  "BL", "BE", "BLE", "BNE",
  "FBL", "FBLE", "JR", "JAL"
};

static long long op_count[32];

static unsigned long fcache_tbl[1 << 12];
static unsigned long long fcache_count;
static unsigned long long pfcache_count;

int main(int argc, char *argv[])
{
  char *p;

  if (argc != 4
      || (fpx = fopen(argv[1], "rb")) == NULL
      || (fpr = fopen(argv[2], "rb")) == NULL
      || (fpw = fopen(argv[3], "wb")) == NULL) {
    printf("Usage: simulate execute_file read_file write_file\n");
    return 1;
  }

  memset(mem, 0xcc, sizeof(mem));

  p = (char *)mem + fread(mem, 1, sizeof(mem), fpx);

  do {

    /* ߂擾B */
    inst = mem[pc];

    /* ߂sAvOJE^̑ǉB */
    dpc = 1;
    op_count[op()]++;
    op_behavior[op()]();
    pc += dpc;

    if (fcache_tbl[pc & 0xFFF] != pc) {
      if (dpc != 1) pfcache_count++;
      fcache_count++;
      fcache_tbl[pc & 0xFFF] = pc;
   } 

    if (status != NULL) {
      char c;
      int i, j;
      /* 󋵂\ */
      printf("%s", status);
      printf("----------------------------------------------------------------\n");
      /* I[256oCgɃAZũR}hC}Ă */
      /* R}hCIvVt(Ȃ) */
      sprintf(p, "-d -b %ld", pc);
      /* AZuɕt߂̗lq\ */
      system(p - 256);
      printf("----------------------------------------------------------------\n");
      /* WX^̐l\ */
      for(i = 0; i < 8 ; i++) {
	for (j = 0; j < 4; j++) {
	  const int addr = i | j << 3;
	  printf("r%-2d: ", addr); 
	  printf("%08lX  ", reg[addr]);
	}
	printf("\n");
      }
      printf("----------------------------------------------------------------\n");
      /* s邩ǂ₷ */
      printf("press 'c' to continue step-mode, 'b' to break, 'a' to abort.\n");
      /* 񓚂擾 */
      scanf(" %1s", &c);
      printf("----------------------------------------------------------------\n");
      /* 󋵂XV */
      switch (c) {
      case 'c':
	status = "";
	break;
      case 'b':
	status = NULL;
	break;
      case 'a':
	return 0;
      }
    }
    /* Ȃ΁A~Ɣf */
  } while(dpc);
  {
    int i = -1, j = -1;
    do {
      while (++i < 16 && op_label[i] == NULL);
      if (i >= 16) printf("                                ");
      else printf("%-5s : %22lld  ", op_label[i], op_count[i]);
      while (++j < 16 && fnct_label[j] == NULL);
      if (j >= 16) printf("                                ");
      else printf("%-5s : %22lld  ", fnct_label[j], fnct_count[j]);
      printf("\n");
    } while (i < 16 || j < 16);
  }
  {
    int i;
    unsigned long long count = 0;
    for (i = 0; i < 16; i++) {
      count += op_count[i];
    }
    printf("%-5s : %22lld  ", "TOTAL", count);
    printf("\n");
  }
  printf("%-5s : %22lld  ", "FERR", fcache_count);
  printf("\n");
  printf("%-5s : %22lld  ", "PFERR", pfcache_count);
  printf("\n");
  printf("%-5s : %22lld  ", "LERR", lcache_count);
  printf("\n");

  return 0;
}


